@@ -5,7 +5,7 @@ import importlib
 import inspect
 import json
 import logging
-import os
+import os, stat, errno
 import pickle
 import shutil
 
@@ -17,15 +17,29 @@ from .general import PipelineError
 from .parallel import ParallelMasterContext, ParalelMockMasterContext
 from .progress import ProgressContext
 
+# The following two functions extend shutil.rmtree, because by default it
+# refuses to delete write-protected files on Windows. However, we often want
+# to delete .git directories, which are protected.
+def handle_rmtree_error(delegate, path, exec):
+  if delegate in (os.rmdir, os.remove) and exec[1].errno == errno.EACCES:
+      os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
+      delegate(path)
+  else:
+      raise
+
+def rmtree(path):
+    return shutil.rmtree(path, ignore_errors = False, onerror = handle_rmtree_error)
+
 
 class NoDefaultValue:
     pass
 
 
 class StageInstance:
-    def __init__(self, instance, name):
+    def __init__(self, instance, name, module_hash):
         self.instance = instance
         self.name = name
+        self.module_hash = module_hash
 
     def parameterize(self, parameters):
         return ParameterizedStage(self.instance, self.name)
@@ -46,36 +60,73 @@ class StageInstance:
         else:
             raise RuntimeError("Stage %s does not have execute method" % self.name)
 
+def get_module_hash(module):
+    if os.path.exists(module.__file__):
+        with open(module.__file__) as f:
+            hash = hashlib.md5()
+            hash.update(f.read().encode("ascii"))
+            return hash.hexdigest()
+
+    return None
 
 def resolve_stage(descriptor):
+    module_hash = None
+
     if isinstance(descriptor, str):
         try:
             # Try to get the module referenced by the string
             descriptor = importlib.import_module(descriptor)
+            module_hash = get_module_hash(descriptor)
         except ModuleNotFoundError:
             # Not a module, but maybe a class?
             parts = descriptor.split(".")
 
             module = importlib.import_module(".".join(parts[:-1]))
+            module_hash = get_module_hash(module)
+
             constructor = getattr(module, parts[-1])
             descriptor = constructor()
 
     if inspect.ismodule(descriptor):
-        return StageInstance(descriptor, descriptor.__name__)
+        module_hash = get_module_hash(descriptor)
+        return StageInstance(descriptor, descriptor.__name__, module_hash)
 
     if inspect.isclass(descriptor):
-        return StageInstance(descriptor(), "%s.%s" % (descriptor.__module__, descriptor.__name__))
+        module = importlib.import_module(descriptor.__module__)
+        module_hash = get_module_hash(module)
+        return StageInstance(descriptor(), "%s.%s" % (descriptor.__module__, descriptor.__name__), module_hash)
 
     clazz = descriptor.__class__
-    return StageInstance(descriptor, "%s.%s" % (clazz.__module__, clazz.__name__))
+    return StageInstance(descriptor, "%s.%s" % (clazz.__module__, clazz.__name__), module_hash)
 
+def get_config_path(name, config):
+    if name in config:
+        return name
+
+def has_config_value(name, config):
+    for segment in name.split("."):
+        if segment in config:
+            config = config[segment]
+        else:
+            return False
+
+    return True
+
+def get_config_value(name, config):
+    for segment in name.split("."):
+        if segment in config:
+            config = config[segment]
+        else:
+            return None
+
+    return config
 
 def configure_stage(instance, context, config):
     config_values = {}
 
     for name, default_value in context.required_config.items():
-        if name in config:
-            config_values[name] = config[name]
+        if name in has_config_value(name, config):
+            config_values[name] = get_config_value(name, config)
         elif not type(default_value) == NoDefaultValue:
             config_values[name] = default_value
         else:
@@ -129,8 +180,8 @@ class ConfigurationContext:
         self.ephemeral_mask = []
 
     def config(self, option, default = NoDefaultValue()):
-        if option in self.base_config:
-            self.required_config[option] = self.base_config[option]
+        if has_config_value(option, self.base_config):
+            self.required_config[option] = get_config_value(option, self.base_config)
         elif not isinstance(default, NoDefaultValue):
             if option in self.required_config and not self.required_config[option] == default:
                 raise PipelineError("Got multiple default values for config option: %s" % option)
@@ -440,6 +491,8 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
         if not os.path.isdir(working_directory):
             raise PipelineError("Working directory does not exist: %s" % working_directory)
 
+        working_directory = os.path.realpath(working_directory)
+
     # 1) Construct stage registry
     registry = process_stages(definitions, config)
 
@@ -539,6 +592,19 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
     # 4.3) Devalidate if configuration values have changed
     # This devalidation step is obsolete since we have implicit config parameters
 
+    # 4.3) Devalidate if module hash of a stage has changed
+    for hash in sorted_cached_hashes:
+        if hash in meta:
+            if not "module_hash" in meta[hash]:
+                stale_hashes.add(hash) # Backwards compatibility
+
+            else:
+                previous_module_hash = meta[hash]["module_hash"]
+                current_module_hash = registry[hash]["wrapper"].module_hash
+
+                if previous_module_hash != current_module_hash:
+                    stale_hashes.add(hash)
+
     # 4.3) Devalidate if cache is not existant
     if not working_directory is None:
         for hash in sorted_cached_hashes:
@@ -640,7 +706,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
 
             if not working_directory is None:
                 if os.path.exists(cache_path):
-                    shutil.rmtree(cache_path)
+                    rmtree(cache_path)
                 os.mkdir(cache_path)
 
             context = ExecuteContext(stage["config"], stage["required_stages"], stage["aliases"], working_directory, stage["dependencies"], cache_path, pipeline_config, logger, cache, stage_dependency_info)
@@ -665,7 +731,8 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
                     dependency_hash: meta[dependency_hash]["updated"] for dependency_hash in stage["dependencies"]
                 },
                 "info": context.stage_info,
-                "validation_token": validation_token
+                "validation_token": validation_token,
+                "module_hash": stage["wrapper"].module_hash
             }
 
             if not working_directory is None:
@@ -681,7 +748,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
                             cache_directory_path = "%s/%s.cache" % (working_directory, dependency_hash)
                             cache_file_path = "%s/%s.p" % (working_directory, dependency_hash)
 
-                            shutil.rmtree(cache_directory_path)
+                            rmtree(cache_directory_path)
                             os.remove(cache_file_path)
 
                             logger.info("Removed ephemeral %s." % dependency_hash)
